(* Copyright (c) 2008 Gary A. Huber, Howard Hughes Medical Institute 
   See the file COPYRIGHT for copying permission
*)


(* 
Used internally.  These are useful functions for 
getting information from XML parse trees generated by
the functions in the Jam_xml_ml_pull_parser module.

The "child" node in the function names is denoted
by the second, string argument.

*)

module JP = Jam_xml_ml_pull_parser
type node = JP.node;;

exception Not_found of string;;

let checked_child node tag = 
  let nopt = JP.child node tag in
    match nopt with
      | None -> raise (Not_found tag)
      | Some node -> node

let int_of_child node tag = 
  let cnode = checked_child node tag in
  let stm = JP.node_stream cnode in
    JP.int_from_stream stm

let int64_of_child node tag = 
  let cnode = checked_child node tag in
  let stm = JP.node_stream cnode in
    JP.int64_from_stream stm

let ints_of_children node tag = 
  let cnodes = JP.children node tag in
    List.map 
      (fun cnode ->
	let stm = JP.node_stream cnode in
	  JP.int_from_stream stm
      )
      cnodes

let float_of_child node tag = 
  let cnode = checked_child node tag in
  let stm = JP.node_stream cnode in
    JP.float_from_stream stm

let float3_of_child node tag =
  let cnode = checked_child node tag in
  let stm = JP.node_stream cnode in
  let x = JP.float_from_stream stm in
  let y = JP.float_from_stream stm in
  let z = JP.float_from_stream stm in
    x,y,z

let floats_of_children node tag = 
  let cnodes = JP.children node tag in
    List.map 
      (fun cnode ->
	let stm = JP.node_stream cnode in
	  JP.float_from_stream stm
      )
      cnodes

let string_of_child node tag = 
  let cnode = checked_child node tag in
  let stm = JP.node_stream cnode in
    JP.string_from_stream stm



